/*
A port of (https://github.com/hypernewbie/VEFontCache) to Odin.

See: https://github.com/Ed94/VEFontCache-Odin
*/
package vefontcache

import "base:runtime"

// White: Cached Hit, Red: Cache Miss, Yellow: Oversized
ENABLE_DRAW_TYPE_VIS :: false
// See: mappings.odin for profiling hookup
DISABLE_PROFILING    :: true

Font_ID :: distinct i32
Glyph   :: distinct i32

Load_Font_Error :: enum(i32) {
	None,
	Parser_Failed,
}

Entry :: struct {
	parser_info    : Parser_Font_Info,
	shaper_info    : Shaper_Info,
	id             : Font_ID,
	used           : b32,
	curve_quality  : f32,
	// snap_glyph_pos : 

	ascent   : f32,
	descent  : f32,
	line_gap : f32,
}

Entry_Default :: Entry {
	id            = 0,
	used          = false,
	curve_quality = 3,
}

Context :: struct {
	backing : Allocator,

	parser_ctx  : Parser_Context, // Glyph parser state
	shaper_ctx  : Shaper_Context, // Text shaper state

	// The managed font instances
	entries : [dynamic]Entry,

	glyph_buffer : Glyph_Draw_Buffer,
	atlas        : Atlas,
	shape_cache  : Shaped_Text_Cache,
	draw_list    : Draw_List,

	// Tracks the offsets for the current layer in a draw_list
	draw_layer : struct {
		vertices_offset : int,
		indices_offset  : int,
		calls_offset    : int,
	},

	debug_print         : b32,
	debug_print_verbose : b32,

	//TODO(Ed):  Add a push/pop stack for the below

	// Helps with hinting
	snap_width  : f32,
	snap_height : f32,

	// camera        : Camera, // TODO(Ed): Add camera support
	colour        : Colour, // Color used in draw interface
	cursor_pos    : Vec2,
	alpha_sharpen : f32,    // Will apply a boost scalar (1.0 + alpha sharpen) to the colour's alpha which provides some sharpening of the edges.
	// Used by draw interface to super-scale the text by 
	// upscaling px_size with px_scalar and then down-scaling
	// the draw_list result by the same amount.
	px_scalar     : f32,   // Used to upscale which font size is used to render (improves hinting)

	default_curve_quality : i32,
}

Init_Atlas_Params :: struct {
	size_multiplier : u32, // How much to scale the the atlas size to. (Affects everything, the base is 4096 x 2048 and everything follows from there)
	glyph_padding   : u32, // Padding to add to bounds_<width/height>_scaled for choosing which atlas region.
}

Init_Atlas_Params_Default :: Init_Atlas_Params {
	size_multiplier = 1,
	glyph_padding   = 1,
}

Init_Glyph_Draw_Params :: struct {
	over_sample               : u32,
	draw_padding              : u32,
	shape_gen_scratch_reserve : u32,
	buffer_glyph_limit        : u32, // How many region.D glyphs can be drawn to the glyph render target buffer at once (worst case scenario)
	batch_glyph_limit         : u32, // How many glyphs can at maximimum be proccessed at once by batch_generate_glyphs_draw_list
}

Init_Glyph_Draw_Params_Default :: Init_Glyph_Draw_Params {
	over_sample                     = 4,
	draw_padding                    = Init_Atlas_Params_Default.glyph_padding,
	shape_gen_scratch_reserve       = 10 * 1024,
	buffer_glyph_limit              = 4,
	batch_glyph_limit               = 512,
}

Init_Shaper_Params :: struct {
	snap_glyph_position           : b32,
	adv_snap_small_font_threshold : u32,
}

Init_Shaper_Params_Default :: Init_Shaper_Params {
	snap_glyph_position           = false,
	adv_snap_small_font_threshold = 0,
}

Init_Shape_Cache_Params :: struct {
	capacity       : u32,
	reserve_length : u32,
}

Init_Shape_Cache_Params_Default :: Init_Shape_Cache_Params {
	capacity       = 10 * 1024,
	reserve_length = 10 * 1024,
}

//#region("lifetime")

// ve_fontcache_init
startup :: proc( ctx : ^Context, parser_kind : Parser_Kind = .STB_TrueType,
	allocator                   := context.allocator,
	atlas_params                := Init_Atlas_Params_Default,
	glyph_draw_params           := Init_Glyph_Draw_Params_Default,
	shape_cache_params          := Init_Shape_Cache_Params_Default,
	shaper_params               := Init_Shaper_Params_Default,
	alpha_sharpen               : f32 = 0.2,
	px_scalar                   : f32 = 1.0,
	
	// Curve quality to use for a font when unspecified,
	// Affects step size for bezier curve passes in generate_glyph_pass_draw_list
	default_curve_quality       : u32 = 3,
	entires_reserve             : u32 = 256,
)
{
	assert( ctx != nil, "Must provide a valid context" )

	ctx.backing       = allocator
	context.allocator = ctx.backing

	ctx.colour        = { 1, 1, 1, 1 }
	ctx.alpha_sharpen = alpha_sharpen
	ctx.px_scalar     = px_scalar

	shaper_ctx := & ctx.shaper_ctx
	shaper_ctx.adv_snap_small_font_threshold = f32(shaper_params.adv_snap_small_font_threshold)
	shaper_ctx.snap_glyph_position           = shaper_params.snap_glyph_position

	ctx.default_curve_quality = default_curve_quality == 0 ? 3 : i32(default_curve_quality)

	error : Allocator_Error
	ctx.entries, error = make( [dynamic]Entry, len = 0, cap = entires_reserve )
	assert(error == .None, "VEFontCache.init : Failed to allocate entries")

	ctx.draw_list.vertices, error = make( [dynamic]Vertex, len = 0, cap = 8 * Kilobyte )
	assert(error == .None, "VEFontCache.init : Failed to allocate draw_list.vertices")

	ctx.draw_list.indices, error = make( [dynamic]u32, len = 0, cap = 16 * Kilobyte )
	assert(error == .None, "VEFontCache.init : Failed to allocate draw_list.indices")

	ctx.draw_list.calls, error = make( [dynamic]Draw_Call, len = 0, cap = Kilobyte )
	assert(error == .None, "VEFontCache.init : Failed to allocate draw_list.calls")
	
	atlas := & ctx.atlas
	Atlas_Setup:
	{
		atlas.size_multiplier = f32(atlas_params.size_multiplier)

		atlas_size    := Vec2i { 4096, 2048 } * i32(atlas.size_multiplier)
		slot_region_a := Vec2i {  32,  32 }   * i32(atlas.size_multiplier)
		slot_region_c := Vec2i {  64,  64 }   * i32(atlas.size_multiplier)
		slot_region_b := Vec2i {  32,  64 }   * i32(atlas.size_multiplier)
		slot_region_d := Vec2i { 128, 128 }   * i32(atlas.size_multiplier)

		init_atlas_region :: proc( region : ^Atlas_Region, atlas_size, slot_size : Vec2i, factor : Vec2i )
		{
			region.next_idx  = 0;
			region.slot_size = slot_size
			region.size      =  atlas_size / factor
			region.capacity  = region.size / region.slot_size

			error : Allocator_Error
			lru_init( & region.state, region.capacity.x * region.capacity.y )
		}
		init_atlas_region( & atlas.region_a, atlas_size, slot_region_a, { 4, 2})
		init_atlas_region( & atlas.region_b, atlas_size, slot_region_b, { 4, 2})
		init_atlas_region( & atlas.region_c, atlas_size, slot_region_c, { 4, 1})
		init_atlas_region( & atlas.region_d, atlas_size, slot_region_d, { 2, 1})

		atlas.size          = atlas_size
		atlas.glyph_padding = f32(atlas_params.glyph_padding)

		atlas.region_a.offset   = {0, 0}
		atlas.region_b.offset.x = 0
		atlas.region_b.offset.y = atlas.region_a.size.y
		atlas.region_c.offset.x = atlas.region_a.size.x
		atlas.region_c.offset.y = 0
		atlas.region_d.offset.x = atlas.size.x / 2
		atlas.region_d.offset.y = 0

		atlas.regions = {
			nil,
			& atlas.region_a,
			& atlas.region_b,
			& atlas.region_c,
			& atlas.region_d,
		}
	}

	Shape_Cache_Setup:
	{
		shape_cache := & ctx.shape_cache
		lru_init( & shape_cache.state, i32(shape_cache_params.capacity) )

		shape_cache.storage, error = make( [dynamic]Shaped_Text, shape_cache_params.capacity )
		assert(error == .None, "VEFontCache.init : Failed to allocate shape_cache.storage")

		for idx : u32 = 0; idx < shape_cache_params.capacity; idx += 1 {
			stroage_entry := & shape_cache.storage[idx]

			stroage_entry.glyphs, error = make( [dynamic]Glyph, len = 0, cap = shape_cache_params.reserve_length )
			assert( error == .None, "VEFontCache.init : Failed to allocate glyphs array for shape cache storage" )

			stroage_entry.positions, error = make( [dynamic]Vec2, len = 0, cap = shape_cache_params.reserve_length )
			assert( error == .None, "VEFontCache.init : Failed to allocate positions array for shape cache storage" )
		}
	}

	Glyph_Buffer_Setup:
	{
		glyph_buffer := & ctx.glyph_buffer
		glyph_buffer.over_sample   = { f32(glyph_draw_params.over_sample), f32(glyph_draw_params.over_sample) }
		glyph_buffer.size.x        = atlas.region_d.slot_size.x * i32(glyph_buffer.over_sample.x) * i32(glyph_draw_params.buffer_glyph_limit)
		glyph_buffer.size.y        = atlas.region_d.slot_size.y * i32(glyph_buffer.over_sample.y)
		glyph_buffer.draw_padding  = cast(f32) glyph_draw_params.draw_padding

		buffer_limit := glyph_draw_params.buffer_glyph_limit
		batch_limit  := glyph_draw_params.batch_glyph_limit

		glyph_buffer.draw_list.vertices, error = make( [dynamic]Vertex, len = 0, cap = 8 * Kilobyte )
		assert( error == .None, "VEFontCache.init : Failed to allocate vertices array for glyph_buffer.draw_list" )

		glyph_buffer.draw_list.indices, error = make( [dynamic]u32, len = 0, cap = 16 * Kilobyte )
		assert( error == .None, "VEFontCache.init : Failed to allocate indices for glyph_buffer.draw_list" )

		glyph_buffer.draw_list.calls, error = make( [dynamic]Draw_Call, len = 0, cap = Kilobyte )
		assert( error == .None, "VEFontCache.init : Failed to allocate calls for glyph_buffer.draw_list" )

		glyph_buffer.clear_draw_list.vertices, error = make( [dynamic]Vertex, len = 0, cap = 2 * Kilobyte )
		assert( error == .None, "VEFontCache.init : Failed to allocate vertices array for clear_draw_list" )

		glyph_buffer.clear_draw_list.indices, error = make( [dynamic]u32, len = 0, cap = 4 * Kilobyte )
		assert( error == .None, "VEFontCache.init : Failed to allocate calls for indices array for clear_draw_list" )

		glyph_buffer.clear_draw_list.calls, error = make( [dynamic]Draw_Call, len = 0, cap = Kilobyte )
		assert( error == .None, "VEFontCache.init : Failed to allocate calls for calls for clear_draw_list" )

		glyph_buffer.shape_gen_scratch, error = make( [dynamic]Vertex, len = 0, cap = 4 * Kilobyte )
		assert(error == .None, "VEFontCache.init : Failed to allocate shape_gen_scratch")

		batch_cache    := & glyph_buffer.batch_cache
		batch_cache.cap = i32(glyph_draw_params.batch_glyph_limit)
		batch_cache.num = 0
		batch_cache.table, error = make( map[Atlas_Region_Key]b8, uint(glyph_draw_params.shape_gen_scratch_reserve) )
		assert(error == .None, "VEFontCache.init : Failed to allocate batch_cache")

		glyph_buffer.glyph_pack,error = make_soa( #soa[dynamic]Glyph_Pack_Entry, length = 0, capacity = 1 * Kilobyte )
		glyph_buffer.oversized, error = make( [dynamic]i32, len = 0, cap = 1 * Kilobyte )
		glyph_buffer.to_cache,  error = make( [dynamic]i32, len = 0, cap = 1 * Kilobyte )
		glyph_buffer.cached,    error = make( [dynamic]i32, len = 0, cap = 1 * Kilobyte )
	}

	parser_init( & ctx.parser_ctx, parser_kind )
	shaper_init( & ctx.shaper_ctx )
}

hot_reload :: proc( ctx : ^Context, allocator : Allocator )
{
	assert( ctx != nil )
	ctx.backing       = allocator
	context.allocator = ctx.backing

	atlas        := & ctx.atlas
	glyph_buffer := & ctx.glyph_buffer
	shape_cache  := & ctx.shape_cache
	draw_list    := & ctx.draw_list

	reload_array( & ctx.entries, allocator )

	reload_array( & glyph_buffer.draw_list.calls,    allocator )
	reload_array( & glyph_buffer.draw_list.indices,  allocator )
	reload_array( & glyph_buffer.draw_list.vertices, allocator )

	reload_array( & glyph_buffer.clear_draw_list.calls,    allocator )
	reload_array( & glyph_buffer.clear_draw_list.indices,  allocator )
	reload_array( & glyph_buffer.clear_draw_list.vertices, allocator )

	reload_map(   & glyph_buffer.batch_cache.table, allocator )
	reload_array( & glyph_buffer.shape_gen_scratch, allocator )

	reload_array_soa( & glyph_buffer.glyph_pack, allocator )
	reload_array(     & glyph_buffer.oversized,  allocator )
	reload_array(     & glyph_buffer.to_cache,   allocator )
	reload_array(     & glyph_buffer.cached,     allocator )

	lru_reload( & atlas.region_a.state, allocator)
	lru_reload( & atlas.region_b.state, allocator)
	lru_reload( & atlas.region_c.state, allocator)
	lru_reload( & atlas.region_d.state, allocator)

	lru_reload( & shape_cache.state, allocator )
	for idx : i32 = 0; idx < i32(len(shape_cache.storage)); idx += 1 {
		storage_entry := & shape_cache.storage[idx]
		reload_array( & storage_entry.glyphs,    allocator )
		reload_array( & storage_entry.positions, allocator )
	}
	reload_array( & shape_cache.storage, allocator )
	
	reload_array( & draw_list.vertices, allocator)
	reload_array( & draw_list.indices,  allocator )
	reload_array( & draw_list.calls,    allocator )
}

shutdown :: proc( ctx : ^Context )
{
	assert( ctx != nil )
	context.allocator = ctx.backing

	atlas        := & ctx.atlas
	glyph_buffer := & ctx.glyph_buffer
	shape_cache  := & ctx.shape_cache
	draw_list    := & ctx.draw_list

	for & entry in ctx.entries {
		unload_font( ctx, entry.id )
	}
	delete( ctx.entries )
	
	delete( glyph_buffer.draw_list.vertices )
	delete( glyph_buffer.draw_list.indices )
	delete( glyph_buffer.draw_list.calls )

	delete( glyph_buffer.clear_draw_list.vertices )
	delete( glyph_buffer.clear_draw_list.indices )
	delete( glyph_buffer.clear_draw_list.calls )

	delete( glyph_buffer.batch_cache.table )
	delete( glyph_buffer.shape_gen_scratch )

	delete_soa( glyph_buffer.glyph_pack)
	delete(     glyph_buffer.oversized)
	delete(     glyph_buffer.to_cache)
	delete(     glyph_buffer.cached)

	lru_free( & atlas.region_a.state )
	lru_free( & atlas.region_b.state )
	lru_free( & atlas.region_c.state )
	lru_free( & atlas.region_d.state )

	for idx : i32 = 0; idx < i32(len(shape_cache.storage)); idx += 1 {
		storage_entry := & shape_cache.storage[idx]
		delete( storage_entry.glyphs )
		delete( storage_entry.positions )
	}
	lru_free( & shape_cache.state )
	
	delete( draw_list.vertices )
	delete( draw_list.indices )
	delete( draw_list.calls )

	shaper_shutdown( & ctx.shaper_ctx )
	parser_shutdown( & ctx.parser_ctx )
}

load_font :: proc( ctx : ^Context, label : string, data : []byte, glyph_curve_quality : u32 = 0 ) -> (font_id : Font_ID, error : Load_Font_Error)
{
	profile(#procedure)
	assert( ctx != nil )
	assert( len(data) > 0 )
	context.allocator = ctx.backing

	entries := & ctx.entries

	id : i32 = -1

	for index : i32 = 0; index < i32(len(entries)); index += 1 {
		if entries[index].used do continue
		id = index
		break
	}
	if id == -1 {
		append_elem( entries, Entry {})
		id = cast(i32) len(entries) - 1
	}
	assert( id >= 0 && id < i32(len(entries)) )

	entry := & entries[ id ]
	{
		entry.used = true

		profile_begin("calling loaders")
		parser_error : b32
		entry.parser_info, parser_error = parser_load_font( & ctx.parser_ctx, label, data )
		if parser_error {
			error = .Parser_Failed
			return
		}
		entry.shaper_info = shaper_load_font( & ctx.shaper_ctx, label, data )
		profile_end()

		ascent, descent, line_gap := parser_get_font_vertical_metrics(entry.parser_info)
		entry.ascent   = f32(ascent)
		entry.descent  = f32(descent)
		entry.line_gap = f32(line_gap)

		if glyph_curve_quality == 0 {
			entry.curve_quality = f32(ctx.default_curve_quality)
		}
		else {
			entry.curve_quality = f32(glyph_curve_quality)
		}
	}
	entry.id = Font_ID(id)
	ctx.entries[ id ].id = Font_ID(id)

	font_id = Font_ID(id)
	return
}

unload_font :: proc( ctx : ^Context, font : Font_ID )
{
	assert( ctx != nil )
	assert( font >= 0 && int(font) < len(ctx.entries) )
	context.allocator = ctx.backing

	entry     := & ctx.entries[ font ]
	entry.used = false

	parser_unload_font( & entry.parser_info )
	shaper_unload_font( & entry.shaper_info )
}

//#endregion("lifetime")

//#region("drawing")

configure_snap :: #force_inline proc( ctx : ^Context, snap_width, snap_height : u32 ) {
	assert( ctx != nil )
	ctx.snap_width  = f32(snap_width)
	ctx.snap_height = f32(snap_height)
}

get_cursor_pos     :: #force_inline proc( ctx : ^Context                  ) -> Vec2 { assert(ctx != nil); return ctx.cursor_pos      }
set_alpha_scalar   :: #force_inline proc( ctx : ^Context, scalar : f32    )         { assert(ctx != nil); ctx.alpha_sharpen = scalar }
set_px_scalar      :: #force_inline proc( ctx : ^Context, scalar : f32    )         { assert(ctx != nil); ctx.px_scalar     = scalar } 
set_colour         :: #force_inline proc( ctx : ^Context, colour : Colour )         { assert(ctx != nil); ctx.colour        = colour }

set_snap_glyph_pos :: #force_inline proc( ctx : ^Context, should_snap : b32 ) { 
	assert(ctx != nil)
	ctx.shaper_ctx.snap_glyph_position = should_snap
}

@(optimization_mode="favor_size")
draw_text :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, position, scale : Vec2, text_utf8 : string )
{
	profile(#procedure)
	assert( ctx != nil )
	assert( font >= 0 && int(font) < len(ctx.entries) )
	assert( len(text_utf8) > 0 )

	entry := ctx.entries[ font ]

	ctx.cursor_pos = {}

	position := position
	position.x = ceil(position.x * ctx.snap_width ) / ctx.snap_width
	position.y = ceil(position.y * ctx.snap_height) / ctx.snap_height

	colour   := ctx.colour
	colour.a  = 1.0 + ctx.alpha_sharpen

	font_scale := parser_scale( entry.parser_info, px_size )

	px_upscale         := px_size * ctx.px_scalar
	downscale          := scale * (1 / ctx.px_scalar)
	font_scale_upscale := parser_scale( entry.parser_info, px_upscale )

	shape := shaper_shape_text_cached( text_utf8, & ctx.shaper_ctx, & ctx.shape_cache, 
		font, 
		entry, 
		px_upscale,
		font_scale_upscale, 
		shaper_shape_text_uncached_advanced
	)
	ctx.cursor_pos = generate_shape_draw_list( & ctx.draw_list, shape, & ctx.atlas, & ctx.glyph_buffer, ctx.px_scalar,
		colour, 
		entry, 
		px_upscale,
		font_scale_upscale, 
		position,
		downscale, 
		ctx.snap_width, 
		ctx.snap_height
	)
}

draw_text_slice :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, position, scale : Vec2, texts : []string )
{
	profile(#procedure)
	assert( ctx != nil )
	assert( font >= 0 && int(font) < len(ctx.entries) )
	assert( len(texts) > 0 )

	entry := ctx.entries[ font ]

	ctx.cursor_pos = {}

	position := position
	position.x = ceil(position.x * ctx.snap_width ) / ctx.snap_width
	position.y = ceil(position.y * ctx.snap_height) / ctx.snap_height

	colour   := ctx.colour
	colour.a  = 1.0 + ctx.alpha_sharpen

	font_scale := parser_scale( entry.parser_info, px_size )

	px_upscale         := px_size * ctx.px_scalar
	downscale          := scale * (1 / ctx.px_scalar)
	font_scale_upscale := parser_scale( entry.parser_info, px_upscale )

	shapes := make( []Shaped_Text, len(texts) )
	for str, id in texts {
		assert( len(str) > 0 )
		shape := shaper_shape_text_cached( str, & ctx.shaper_ctx, & ctx.shape_cache, 
			font, 
			entry,
			px_upscale,
			font_scale_upscale, 
			shaper_shape_text_uncached_advanced
		)
		shapes[id] = shape
	}

	generate_shapes_draw_list(ctx, font, colour, entry, px_upscale, font_scale_upscale, position, downscale, shapes )
}


// draw_text_no_snap :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, position, scale : Vec2, text_utf8 : string )
// {
// 	profile(#procedure)
// 	assert( ctx != nil )
// 	assert( font >= 0 && int(font) < len(ctx.entries) )
// 	assert( len(text_utf8) > 0 )

// 	ctx.cursor_pos = {}

// 	entry := ctx.entries[ font ]

// 	colour   := ctx.colour
// 	colour.a  = 1.0 + ctx.alpha_sharpen

// 	px_size_upscaled   := px_size * ctx.px_scalar
// 	scale_downsample   := scale * (1 / ctx.px_scalar)
// 	font_scale_upscale := parser_scale( entry.parser_info, px_size_upscaled )

// 	font_scale    := parser_scale( entry.parser_info, -px_size )
// 	shape         := shaper_shape_text_cached( text_utf8, & ctx.shaper_ctx, & ctx.shape_cache, font, entry, px_size, font_scale, shaper_shape_text_latin )
// 	ctx.cursor_pos = generate_shape_draw_list( & ctx.draw_list, shape, & ctx.atlas, & ctx.glyph_buffer, colour, entry, font_scale, position, scale, ctx.snap_width, ctx.snap_height )
// }

// Resolve the shape and track it to reduce iteration overhead
@(optimization_mode="favor_size")
draw_text_shape :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, position, scale : Vec2, shape : Shaped_Text )
{
	profile(#procedure)
	assert( ctx != nil )
	assert( font >= 0 && int(font) < len(ctx.entries) )
	position := position
	position.x = ceil(position.x * ctx.snap_width ) / ctx.snap_width
	position.y = ceil(position.y * ctx.snap_height) / ctx.snap_height

	entry := ctx.entries[ font ]

	colour   := ctx.colour
	colour.a  = 1.0 + ctx.alpha_sharpen

	font_scale := parser_scale( entry.parser_info, px_size )

	px_upscale         := px_size * ctx.px_scalar
	downscale          := scale * (1 / ctx.px_scalar)
	font_scale_upscale := parser_scale( entry.parser_info, px_upscale )

	ctx.cursor_pos = generate_shape_draw_list( & ctx.draw_list, shape, & ctx.atlas, & ctx.glyph_buffer, ctx.px_scalar,
		colour, 
		entry, 
		px_upscale,
		font_scale_upscale, 
		position, 
		downscale, 
		ctx.snap_width, 
		ctx.snap_height
	)
}

// Resolve the shape and track it to reduce iteration overhead
// draw_text_shape_no_snap :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, position, scale : Vec2, shape : Shaped_Text )
// {
// 	profile(#procedure)
// 	assert( ctx != nil )
// 	assert( font >= 0 && int(font) < len(ctx.entries) )

// 	colour   := ctx.colour
// 	colour.a  = 1.0 + ctx.alpha_sharpen

// 	px_size_upscaled := px_size * ctx.px_scalar
// 	scale            := scale * (1 / ctx.px_scalar)

// 	entry := ctx.entries[ font ]
// 	font_scale     := parser_scale( entry.parser_info, px_size )
// 	ctx.cursor_pos  = generate_shape_draw_list( & ctx.draw_list, shape, & ctx.atlas, & ctx.glyph_buffer, colour, entry, font_scale, position, scale, ctx.snap_width, ctx.snap_height )
// }

get_draw_list :: #force_inline proc( ctx : ^Context, optimize_before_returning := true ) -> ^Draw_List {
	assert( ctx != nil )
	if optimize_before_returning do optimize_draw_list( & ctx.draw_list, 0 )
	return & ctx.draw_list
}

get_draw_list_layer :: #force_inline proc( ctx : ^Context, optimize_before_returning := true ) -> (vertices : []Vertex, indices : []u32, calls : []Draw_Call) {
	assert( ctx != nil )
	if optimize_before_returning do optimize_draw_list( & ctx.draw_list, ctx.draw_layer.calls_offset )
	vertices = ctx.draw_list.vertices[ ctx.draw_layer.vertices_offset : ]
	indices  = ctx.draw_list.indices [ ctx.draw_layer.indices_offset  : ]
	calls    = ctx.draw_list.calls   [ ctx.draw_layer.calls_offset    : ]
	return
}

flush_draw_list :: #force_inline proc( ctx : ^Context ) {
	assert( ctx != nil )
	clear_draw_list( & ctx.draw_list )
	ctx.draw_layer.vertices_offset = 0
	ctx.draw_layer.indices_offset  = 0
	ctx.draw_layer.calls_offset    = 0
}

flush_draw_list_layer :: #force_inline proc( ctx : ^Context ) {
	assert( ctx != nil )
	ctx.draw_layer.vertices_offset = len(ctx.draw_list.vertices)
	ctx.draw_layer.indices_offset  = len(ctx.draw_list.indices)
	ctx.draw_layer.calls_offset    = len(ctx.draw_list.calls)
}

//#endregion("drawing")

//#region("metrics")

measure_text_size :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, text_utf8 : string ) -> (measured : Vec2)
{
	// profile(#procedure)
	assert( ctx != nil )
	assert( font >= 0 && int(font) < len(ctx.entries) )

	entry := ctx.entries[font]

	font_scale := parser_scale( entry.parser_info, px_size )
	
	px_size_upscaled    := px_size * ctx.px_scalar
	font_scale_upscaled := parser_scale( entry.parser_info, px_size_upscaled )

	shaped     := shaper_shape_text_cached( text_utf8, & ctx.shaper_ctx, & ctx.shape_cache, font, entry, px_size_upscaled, font_scale_upscaled, shaper_shape_text_uncached_advanced )
	return shaped.size
}

get_font_vertical_metrics :: #force_inline proc ( ctx : ^Context, font : Font_ID, px_size : f32 ) -> ( ascent, descent, line_gap : f32 )
{
	assert( ctx != nil )
	assert( font >= 0 && int(font) < len(ctx.entries) )

	entry := ctx.entries[ font ]

	font_scale := parser_scale( entry.parser_info, px_size )
	
	px_size_upscaled    := px_size * ctx.px_scalar
	downscale           := 1 / px_size_upscaled
	font_scale_upscaled := parser_scale( entry.parser_info, px_size_upscaled )

	ascent   = font_scale * entry.ascent
	descent  = font_scale * entry.descent
	line_gap = font_scale * entry.line_gap
	return
}

//#endregion("metrics")

//#region("shaping")

shape_text_latin :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, text_utf8 : string ) -> Shaped_Text
{
	profile(#procedure)
	assert( len(text_utf8) > 0 )
	entry := ctx.entries[ font ]

	font_scale := parser_scale( entry.parser_info, px_size )

	px_size_upscaled    := px_size * ctx.px_scalar
	font_scale_upscaled := parser_scale( entry.parser_info, px_size_upscaled )

	return shaper_shape_text_cached( text_utf8, & ctx.shaper_ctx, & ctx.shape_cache, 
		font, 
		entry, 
		px_size, 
		font_scale_upscaled, 
		shaper_shape_text_latin
	)
}

shape_text_advanced :: #force_inline proc( ctx : ^Context, font : Font_ID, px_size : f32, text_utf8 : string ) -> Shaped_Text
{
	profile(#procedure)
	assert( len(text_utf8) > 0 )
	entry := ctx.entries[ font ]

	font_scale := parser_scale( entry.parser_info, px_size )

	px_size_upscaled    := px_size * ctx.px_scalar
	font_scale_upscaled := parser_scale( entry.parser_info, px_size_upscaled )

	return shaper_shape_text_cached( text_utf8, & ctx.shaper_ctx, & ctx.shape_cache,
		font, 
		entry, 
		px_size, 
		font_scale_upscaled, 
		shaper_shape_text_uncached_advanced
	)
}

// User handled shaped text. Will not be cached
shape_text_latin_uncached :: #force_inline proc( ctx : ^Context, font : Font_ID, text_utf8 : string, entry : ^Entry, shape : ^Shaped_Text )
{
	assert(false)
	return
}

// User handled shaped text. Will not be cached
shape_text_advanced_uncahed :: #force_inline proc( ctx : ^Context, font : Font_ID, text_utf8 : string, entry : ^Entry, shape : ^Shaped_Text )
{
	assert(false)
	return
}

//#endregion("shaping")

// Can be used with hot-reload
clear_atlas_region_caches :: proc(ctx : ^Context)
{
	lru_clear(& ctx.atlas.region_a.state)
	lru_clear(& ctx.atlas.region_b.state)
	lru_clear(& ctx.atlas.region_c.state)
	lru_clear(& ctx.atlas.region_d.state)

	ctx.atlas.region_a.next_idx = 0
	ctx.atlas.region_b.next_idx = 0
	ctx.atlas.region_c.next_idx = 0
	ctx.atlas.region_d.next_idx = 0
}

// Can be used with hot-reload
clear_shape_cache :: proc (ctx : ^Context)
{
	lru_clear(& ctx.shape_cache.state)
	for idx : i32 = 0; idx < cast(i32) cap(ctx.shape_cache.storage); idx += 1 {
		stroage_entry := & ctx.shape_cache.storage[idx]
		stroage_entry.end_cursor_pos = {}
		stroage_entry.size           = {}
		clear(& stroage_entry.glyphs)
		clear(& stroage_entry.positions)
	}
	ctx.shape_cache.next_cache_id = 0
}
